"""
Main file for ordinary use
"""

import re
from typing import Iterable, Literal
from types import ModuleType
import importlib

import pandas as pd
import yaml

from . import decoder, utils
from .metadata_reader import (
    defaults,
    metadata,
    original_tables,
    LoadTable,
    Attribute as _Attribute,
    OriginalTable as _OriginalTable,
    StandardTable as _StandardTables,
    Table as _Table,
)
from .data_cleaner import open_and_clean_table


class TableHandler:
    """A class for loading parquet files"""

    def __init__(self, table_name: str, year: int, settings: LoadTable) -> None:
        self.table_name = table_name
        self.year = year
        self.file_name = f"{year}_{table_name}.parquet"
        self.local_path = defaults.processed_data.joinpath(self.file_name)
        self.file_url = f"{defaults.online_dir}/parquet_files/{self.file_name}"
        self.settings = settings

    def read(self) -> pd.DataFrame:
        """Read the parquet file"""
        self.local_path.parent.mkdir(exist_ok=True, parents=True)
        if (self.settings.on_missing == "create") or self.settings.recreate:
            if (not self.local_path.exists()) or self.settings.recreate:
                table = open_and_clean_table(self.table_name, self.year)
                if self.settings.save_created:
                    table.to_parquet(self.local_path)
            else:
                table = self.read_local_file()
        elif self.settings.on_missing == "download" or self.settings.redownload:
            if (not self.local_path.exists()) or self.settings.redownload:
                table = self.download()
                if self.settings.save_downloaded:
                    table.to_parquet(self.local_path)
            else:
                table = self.read_local_file()
        else:
            table = self.read_local_file()

        table.attrs["table_name"] = self.table_name
        table.attrs["year"] = self.year
        return table

    def download(self) -> pd.DataFrame:
        """Download parquet file to memory"""
        return pd.read_parquet(self.file_url)

    def read_local_file(self) -> pd.DataFrame:
        """Load parquet file from local hard drive"""
        return pd.read_parquet(self.local_path)


class Applier:
    def __init__(
        self, table: pd.DataFrame, instructions: list, properties: dict | None = None
    ) -> None:
        self.table = table
        self.properties = properties if properties is not None else {}
        self.modules: dict[str, ModuleType] = {}
        for instruction in instructions:
            method_name, method_input = self.extract_method_name(instruction)
            if method_input is None:
                getattr(self, f"_{method_name}")()
            else:
                getattr(self, f"_{method_name}")(method_input)

    def extract_method_name(self, instruction):
        if isinstance(instruction, str):
            method_name = instruction
            method_input = None
        elif isinstance(instruction, dict):
            method_name, method_input = list(instruction.items())[0]
        else:
            raise TypeError
        return method_name, method_input

    def _add_year(self) -> None:
        self.table["Year"] = self.properties["year"]

    def _add_table_name(self) -> None:
        self.table["Table_Name"] = self.properties["table_name"]

    def _add_weights(self) -> None:
        self.table = WeightAdder(self.table).add_weights()

    def _add_classification(self, method_input: dict) -> None:
        settings = decoder.CommodityDecoderSettings(**method_input)
        self.table = decoder.CommodityDecoder(self.table, settings).add_classification()

    def _add_attribute(self, method_input: dict) -> None:
        settings = decoder.IDDecoderSettings(**method_input)
        self.table = decoder.IDDecoder(self.table, settings).add_attribute()

    def _apply_order(self, method_input: list):
        new_columns = [
            column for column in method_input if column in self.table.columns
        ]
        self.table = self.table[new_columns]

    def _create_column(self, method_input=None) -> None:
        if method_input is None:
            return
        column_name = method_input["name"]
        if method_input["type"] == "numerical":
            expression = method_input["expression"]
            self.__apply_numerical_instruction(column_name, expression)
        elif method_input["type"] == "categorical":
            categories = method_input["categories"]
            self.__apply_categorical_instruction(column_name, categories)

    def __apply_numerical_instruction(self, column_name, expression: int | str) -> None:
        if isinstance(expression, int):
            self.table.loc[:, column_name] = expression
            return
        columns_names = re.split(r"[\+\-\*\/\s\.\(\)]+", expression)
        columns_names = [
            name for name in columns_names if not (name.isnumeric() or (name is None))
        ]
        self.table[column_name] = (
            self.table[columns_names].fillna(0).eval(expression, engine="python")
        )

    def __apply_categorical_instruction(
        self, column_name: str, categories: dict
    ) -> None:
        if column_name in self.table.columns:
            categorical_column = self.table[column_name].copy()
        else:
            categorical_column = pd.Series(index=self.table.index, dtype="category")

        for category, condition in categories.items():
            filt = self.__construct_filter(column_name, condition)
            categorical_column = categorical_column.cat.add_categories([category])
            categorical_column.loc[filt] = category

        self.table[column_name] = categorical_column

    def __construct_filter(self, column_name, condition) -> pd.Series:
        if condition is None:
            filt = self.table.index.to_series()
        elif isinstance(condition, str):
            filt = self.table[column_name] == condition
        elif isinstance(condition, list):
            filt = self.table[column_name].isin(condition)
        elif isinstance(condition, dict):
            filts = []
            for other_column, value in condition.items():
                if isinstance(value, (bool, str)):
                    filts.append(self.table[other_column] == value)
                elif isinstance(value, list):
                    filts.append(self.table[other_column].isin(value))
                else:
                    raise KeyError
            filt_sum = pd.concat(filts, axis="columns").sum(axis="columns")
            filt = filt_sum == len(condition)
        else:
            raise KeyError
        return filt

    def _apply_filter(self, conditions: str | list[str]):
        conditions = [conditions] if isinstance(conditions, str) else conditions
        for condition in conditions:
            self.table = self.table.query(condition)

    def _apply_pandas_function(self, method_input: str | None = None) -> None:
        if method_input is None:
            return
        method_input = "self.table" + method_input
        table = pd.eval(method_input, target=self.table, engine="python")
        assert isinstance(table, pd.DataFrame)
        self.table = table

    def _apply_external_function(self, method_input: str | None = None) -> None:
        if method_input is None:
            return
        module_name, func_name = method_input.rsplit(".", 1)
        self.__load_module(module_name)
        func = getattr(self.modules[module_name], func_name)
        self.table = func(self.table)

    def __load_module(self, module_name: str) -> None:
        if module_name not in self.modules:
            self.modules[module_name] = importlib.import_module(module_name)

    def _join(self, method_input: dict | str | None = None):
        if method_input is None:
            return
        if isinstance(method_input, str):
            table_name = method_input
            columns = ["Year", "ID"]
            years = None
        elif isinstance(method_input, dict):
            table_name = method_input["table_name"]
            columns = method_input["columns"]
            years = method_input.get("year", None)
        else:
            raise TypeError
        years = list(self.table["Year"].unique())
        settings = LoadTable()
        other_table = TableLoader(table_name, years, settings).load()
        self.table = self.table.merge(other_table, on=columns)


class TableLoader:
    def __init__(
        self,
        table_name: str,
        years: int | Iterable[int] | str | None,
        settings: LoadTable,
    ):
        self.table_name = table_name
        self.years = utils.parse_years(years)
        self.settings = settings
        self.schema: dict = metadata.schema.copy()
        if table_name in self.schema:
            self.table_schema: dict = self.schema[table_name]
        else:
            self.table_schema = {}
        self.original_tables_cache: dict[str, pd.DataFrame] = {}

    def load(self) -> pd.DataFrame:
        table_list = []
        for year in self.years:
            table = self._load_table(self.table_name, year)
            table_list.append(table)
        table = pd.concat(table_list)
        if "views" in self.table_schema:
            table.view.views = self.table_schema["views"]
        return table

    def _load_table(self, table_name: str, year: int) -> pd.DataFrame:
        if table_name in original_tables:
            table = self._load_original_table(table_name, year)
        elif self.schema[table_name].get("cache_result", False):
            try:
                table = self.read_cached_table(table_name, year)
            except FileNotFoundError:
                table = self._construct_schema_based_table(table_name, year)
                self.save_cache(table, table_name, year)
        else:
            table = self._construct_schema_based_table(table_name, year)
        return table

    def read_cached_table(
        self,
        table_name: str | None = None,
        year: int | None = None,
    ) -> pd.DataFrame:
        if not self.check_table_dependencies(table_name, year):
            raise FileNotFoundError
        file_name = f"{table_name}_{year}.parquet"
        file_path = defaults.cached_data.joinpath(file_name)
        table = pd.read_parquet(file_path)
        return table

    def extract_dependencies(self, table_name: str, year: int) -> dict:
        table_list = [table_name]
        dependencies: dict[str, dict] = {}
        while len(table_list) > 0:
            table = table_list.pop(0)
            if "table_list" in metadata.schema[table]:
                dependencies[table] = metadata.schema[table]
                upstream_tables = utils.MetadataVersionResolver(
                    metadata.schema[table]["table_list"], year=year
                ).get_version()
                assert isinstance(upstream_tables, list)
                table_list.extend(upstream_tables)
            else:
                file_name = f"{year}_{table}.parquet"
                local_path = defaults.processed_data.joinpath(file_name)
                dependencies[table] = {"size": local_path.stat().st_size}
        return dependencies

    def check_table_dependencies(self, table_name, year) -> bool:
        file_name = f"{table_name}_{year}_metadata.yaml"
        cach_metadata_path = defaults.cached_data.joinpath(file_name)
        with open(cach_metadata_path, encoding="utf-8") as file:
            cach_metadata = yaml.safe_load(file)
        file_dependencies = cach_metadata["dependencies"]
        current_dependencies = self.extract_dependencies(table_name, year)
        return file_dependencies == current_dependencies

    def save_cache(
        self,
        table: pd.DataFrame,
        table_name: str,
        year: int,
    ) -> None:
        defaults.cached_data.mkdir(parents=True, exist_ok=True)
        file_name = f"{table_name}_{year}.parquet"
        file_path = defaults.cached_data.joinpath(file_name)
        file_name = f"{table_name}_{year}_metadata.yaml"
        cach_metadata_path = defaults.cached_data.joinpath(file_name)
        file_metadata = {"dependencies": self.extract_dependencies(table_name, year)}
        with open(cach_metadata_path, mode="w", encoding="utf-8") as file:
            yaml.safe_dump(file_metadata, file)
        table.to_parquet(file_path, index=False)

    def _apply_schema(
        self,
        table: pd.DataFrame,
        table_name: str | None = None,
        year: int | None = None,
    ):
        if "instructions" not in self.schema[table_name]:
            return table

        if (table_name is None) and ("table_name" in table.attrs):
            table_name = table.attrs["table_name"]

        if (year is None) and ("year" in table.attrs):
            year = table.attrs["year"]
        instructions = utils.MetadataVersionResolver(
            self.schema[table_name]["instructions"], year
        ).get_version()
        assert isinstance(instructions, list)
        props = {"year": year, "table_name": table_name}
        table = Applier(table, instructions, props).table
        return table

    def _load_original_table(self, table_name: str, year: int) -> pd.DataFrame:
        if f"{table_name}_{year}" in self.original_tables_cache:
            return self.original_tables_cache[f"{table_name}_{year}"]
        table = TableHandler(table_name, year, self.settings).read()
        if not table.empty and (table_name in self.schema):
            table = self._apply_schema(table, table_name, year)
        self.original_tables_cache[f"{table_name}_{year}"] = table
        return table

    def _construct_schema_based_table(self, table_name: str, year: int) -> pd.DataFrame:
        if table_name not in self.schema:
            raise KeyError(f"Table name {table_name} is not available in schema")
        table_names = self.schema[table_name]["table_list"]
        table_names = utils.MetadataVersionResolver(table_names, year).get_version()
        assert isinstance(table_names, (str, list))

        table_list = self._collect_schema_tables(table_names, year)

        table = pd.concat(table_list)
        table = self._apply_schema(table, table_name, year)
        return table

    def _collect_schema_tables(
        self, table_names: str | list[str], year: int
    ) -> list[pd.DataFrame]:
        table_names = [table_names] if isinstance(table_names, str) else table_names
        table_list = []
        for name in table_names:
            table = self._load_table(name, year)
            if not table.empty:
                table_list.append(table)
        return table_list


class WeightAdder:
    def __init__(
        self,
        table: pd.DataFrame,
        method: Literal["default", "external", "household_info"] = "default",
        year_column_name: str = "Year",
    ) -> None:
        self.table = table
        self.method = method
        self.year_column_name = year_column_name

    def load_weights(self, year) -> pd.Series:
        if self.method == "default":
            if year <= 1395:
                method = "external"
            else:
                method = "household_info"
        elif self.method in ["external", "household_info"]:
            method = self.method
        else:
            raise ValueError("Method is not valid")

        if method == "external":
            weights = self._load_from_external_data(year)
        else:
            weights = self._load_from_household_info(year)
        return weights

    @staticmethod
    def _load_from_household_info(year) -> pd.Series:
        settings = LoadTable()
        loader = TableLoader("household_information", year, settings)
        hh_info = loader.load()
        weights = hh_info.set_index("ID")["Weight"]
        return weights

    @staticmethod
    def _load_from_external_data(year) -> pd.Series:
        weights_path = defaults.external_data.joinpath("weights.parquet")
        if not weights_path.exists():
            defaults.external_data.mkdir(parents=True, exist_ok=True)
            utils.download(
                f"{defaults.online_dir}/external_data/weights.parquet", weights_path
            )
        weights = pd.read_parquet(weights_path)
        weights = weights.loc[(year), "Weight"]
        assert isinstance(weights, pd.Series)
        return weights

    def add_weights(self, **kwargs) -> pd.DataFrame:
        years = decoder.extract_column(self.table, self.year_column_name)
        years = years.drop_duplicates()
        weights_list = []
        for year in years:
            weights_list.append(self.load_weights(year))
        weights = pd.concat(
            weights_list, axis="index", keys=years, names=["Year", "ID"]
        )
        self.table = self.table.join(weights, on=["Year", "ID"], **kwargs)
        return self.table



from typing import Callable
import importlib

import pandas as pd

from .. import utils
from ..metadata_reader import defaults, metadata


class ExternalDataCleaner:
    def __init__(self, name: str) -> None:
        self.name = name
        raw_folder_path = defaults.external_data.joinpath("_raw")
        raw_folder_path.mkdir(exist_ok=True, parents=True)

        self.metadata = metadata.external_data.get(self.name, None)
        extension = self._find_extension()
        if extension is not None:
            self.raw_file_path = raw_folder_path.joinpath(f"{self.name}.{extension}")
        else:
            self.raw_file_path = None

    def load_data(self):
        external_data = [file.stem for file in defaults.external_data.iterdir()]
        if self.name in external_data:
            return pd.read_parquet(
                defaults.external_data.joinpath(f"{self.name}.parquet")
            )
        if self.metadata is None:
            raise ValueError("Metadata is not available for this table")
        if "url" in self.metadata:
            table = self.clean_raw_file()
        elif "from" in self.metadata:
            table = self.collect_and_clean()
        else:
            raise ValueError
        self.save_data(table)
        return table

    def _find_extension(self) -> str | None:
        available_extentions = ["xlsx"]
        extension = self.metadata.get("extension", None)
        url = self.metadata.get("url", None)

        if (extension is None) and (url is not None):
            extension = url.rsplit(".", maxsplit=1)[1]

        if extension is not None:
            assert extension in available_extentions
        return extension

    def download_raw_file(self) -> None:
        url = self.metadata["url"]
        utils.download(url, self.raw_file_path)

    def load_raw_file(self) -> pd.DataFrame:
        assert self.raw_file_path is not None
        if not self.raw_file_path.exists():
            self.download_raw_file()
        if self.raw_file_path.suffix in [".xlsx"]:
            table = pd.read_excel(self.raw_file_path, header=None)
        else:
            raise ValueError("Format not supported yet")
        return table

    def clean_raw_file(self, table: pd.DataFrame | None = None) -> pd.DataFrame:
        if table is None:
            table = self.load_raw_file()
        table = self.cleaning_function(table)
        return table

    def collect_and_clean(self) -> pd.DataFrame:
        table_list = [
            ExternalDataCleaner(table).load_data() for table in self.metadata["from"]
        ]
        table = self.cleaning_function(table_list)
        return table

    @property
    def cleaning_function(
        self,
    ) -> Callable[[pd.DataFrame | list[pd.DataFrame]], pd.DataFrame]:
        cleaning_module = importlib.import_module(
            "hbsir.external_data.cleaning_scripts"
        )
        return getattr(cleaning_module, self.metadata["cleaning_function"])

    def save_data(self, table: pd.DataFrame) -> None:
        table.to_parquet(defaults.external_data.joinpath(f"{self.name}.parquet"))



"""
Metadata version resolution module.

Provides MetadataVersionResolver class to retrieve specific versions 
of metadata based on given year.

Supports both simple integer and keyword-based versioning of metadata,
configured via MetadataVersionSettings.

Classes
-------
MetadataVersionResolver - Resolves and returns metadata versions.

MetadataVersionSettings - Configurable settings for versioning.

Example
-------
resolver = MetadataVersionResolver(metadata, year=1380)
version = resolver.get_version()
"""

from typing import Literal, overload
from dataclasses import dataclass


@dataclass
class MetadataVersionSettings:
    """
    Settings for metadata versioning.

    Specifies the field names and valid range used for versioning metadata.

    Attributes
    ----------
    year_range : tuple(int, int)
        Valid min and max year for version numbers.
        Default (1350, 1450).

    year_keyword : str
        Metadata field keyword for year. Default "year".

    version_keyword : str
        Metadata field keyword for version section. Default "versions".

    category_keyword : str
        Metadata field keyword for categories. Default "categories".
    """

    year_range: tuple[int, int] = (1350, 1450)
    year_keyword: str = "year"
    version_keyword: str = "versions"
    items_keyword: str = "items"
    category_keyword: str = "categories"
    item_key_name: str = "item_key"


default_settings = MetadataVersionSettings()


class MetadataVersionResolver:
    """
    Resolves specific metadata versions from versioned metadata.

    Retrieves the appropriate version of metadata based on the provided
    year and versioning information in the metadata. Supports both
    keyword-based and simple integer versioning of metadata dictionaries.

    Versioning configuration is specified via a MetadataVersionSettings instance.

    Parameters
    ----------
    metadata : dict, list, str, int, None
        Metadata to extract version from. Can contain versioning info.

    year : int, optional
        Year to retrieve metadata for. If not provided, will try to extract
        from metadata.

    settings : MetadataVersionSettings, optional
        Versioning settings. Default settings used if not provided.

    Attributes
    ----------
    metadata : dict, list, str, int, None
        Original input metadata.

    settings : MetadataVersionSettings
        Versioning configuration settings.

    Methods
    -------
    get_version() : Retrieves resolved metadata version for given year.

    _is_versioned() : Checks if input metadata is versioned.

    _retrieve_dictionary_version() : Gets specific version from dictionary.

    _detect_version_type() : Determines version type of dictionary.

    _find_version_number() : Gets latest version number <= given year.
    """

    def __init__(
        self,
        metadata: dict | list | str | int | float | None,
        year: int | None = None,
        settings: MetadataVersionSettings = default_settings,
    ):
        self.metadata = metadata
        self.settings = settings
        if year is not None:
            self.year = year
        elif (isinstance(metadata, dict)) and (settings.year_keyword in metadata):
            self.year = metadata[settings.year_keyword]
        elif self._is_versioned(metadata):
            raise NameError("Versioned metadata requires year parameter.")

    def get_version(self) -> dict | list | str | int | float | None:
        """
        Retrieves metadata version for given year.

        Recursively traverses metadata and returns appropriate
        version based on configured versioning settings and provided year.

        Returns
        -------
        dict | list | str | int | None
            Resolved metadata version for provided year.
        """
        return self._retrive_version(self.metadata)

    @overload
    def _retrive_version(self, element: dict) -> dict | list | str | int | float | None:
        ...

    @overload
    def _retrive_version(self, element: list) -> list:
        ...

    @overload
    def _retrive_version(self, element: str) -> str:
        ...

    @overload
    def _retrive_version(self, element: int) -> int:
        ...

    @overload
    def _retrive_version(self, element: float) -> float:
        ...

    @overload
    def _retrive_version(self, element: None) -> None:
        ...

    def _retrive_version(self, element):
        if (element is None) or isinstance(element, (int, float, str)):
            pass
        elif isinstance(element, list):
            element = [self._retrive_version(value) for value in element]
        elif isinstance(element, dict):
            element = self._retrieve_dictionaty_verion(element)
            if (
                isinstance(element, dict)
                and self._detect_version_type(element) == "not_versioned"
            ):
                element = {
                    key: self._retrive_version(value) for key, value in element.items()
                }
            else:
                element = self._retrive_version(element)
        else:
            raise TypeError(f"Element {element} is not resolved")

        return element

    def _is_versioned(self, element):
        if isinstance(element, (int, str)):
            return False
        if isinstance(element, list):
            return not all(not self._is_versioned(value) for value in element)
        if isinstance(element, dict):
            if self._detect_version_type(element) != "not_versioned":
                return True
            return not all(not self._is_versioned(value) for value in element.values())
        raise TypeError

    def _retrieve_dictionaty_verion(
        self, dictionaty: dict
    ) -> dict | list | str | int | None:
        version_type = self._detect_version_type(dictionaty)
        if version_type == "not_versioned":
            return dictionaty
        if version_type == "keyword_versioned":
            versioned_dictionary = dictionaty[self.settings.version_keyword]
        elif version_type == "simple_versioned":
            versioned_dictionary = dictionaty
        else:
            raise ValueError

        latest_verion_number = self._find_verion_number(versioned_dictionary)
        if latest_verion_number == 0:
            latest_verion = {}
        else:
            latest_verion = versioned_dictionary[latest_verion_number]

        if version_type == "keyword_versioned":
            latest_verion = {} if latest_verion is None else latest_verion
            if isinstance(latest_verion, dict):
                latest_verion.update(
                    {
                        key: value
                        for key, value in dictionaty.items()
                        if (key != self.settings.version_keyword)
                        and (key not in latest_verion)
                    }
                )
        latest_verion = None if latest_verion == {} else latest_verion

        return latest_verion

    def _detect_version_type(
        self, dictionaty: dict
    ) -> Literal["keyword_versioned", "simple_versioned", "not_versioned"]:
        if self.settings.version_keyword in dictionaty:
            return "keyword_versioned"
        for element in dictionaty:
            if (
                not isinstance(element, int)
                or (element < self.settings.year_range[0])
                or (element > self.settings.year_range[1])
            ):
                return "not_versioned"
        return "simple_versioned"

    def _find_verion_number(self, dictionaty: dict):
        selected_version = 0
        for version in dictionaty:
            if version <= self.year:
                selected_version = max(selected_version, version)
        return selected_version


class MetadataCategoryResolver(MetadataVersionResolver):
    # pylint: disable=unsubscriptable-object
    # pylint: disable=unsupported-assignment-operation
    def categorize_metadata(self) -> dict:
        """Categorize metadata dictionary into items list.

        Parses the input metadata to build a categorized list
        of items under the 'items' key.

        Example:

        Before:

        metadata:
          key1: val1
          key2: val2

          items:
            item1:
              shared1: foo
              shared2: bar

              categories:
                1:
                  cat1key: catval1
                2:
                  cat2key: catval2

        After:

          items:
            - shared1: foo
              shared2: bar
              cat1key: catval1
              item_key: item1

            - shared1: foo
              shared2: bar
              cat2key: catval2
              item_key: item1


        Retrieves the latest version of the metadata using get_version().
        Checks that metadata is a dictionary.

        Loops through metadata[items_keyword] accessing each item:
            - Checks if item has 'categories' key, splits into categories
            - Copies over shared keys from item to categories
            - Sets configured key_name in each category
            - Extends item_list with list of categories

        Returns the updated metadata dict with categorized 'items' list.

        Raises
        ------

        TypeError
            If metadata input is not a dictionary.

        Returns
        -------

        dict
            Updated metadata with 'items' list of categorized elements.
        """
        metadata = self.get_version()
        if not isinstance(metadata, dict):
            raise TypeError
        items = []
        for key, item in metadata[self.settings.items_keyword].items():
            item_list = self._get_categories(item)
            for element in item_list:
                element[self.settings.item_key_name] = key.strip("_")
            items.extend(item_list)
        metadata[self.settings.items_keyword] = items
        return metadata

    def _get_categories(self, item: dict) -> list:
        if "categories" not in item:
            categories_list = [item]
        else:
            categories_number = list(item["categories"].keys())
            categories_number.sort()
            categories_list = [
                item["categories"][number] for number in categories_number
            ]
            shared_keys = [key for key in item.keys() if key != "categories"]
            for category in categories_list:
                for key in shared_keys:
                    if key not in category.keys():
                        category[key] = item[key]
        return categories_list



import pandas as pd

from .external_data_cleaner import ExternalDataCleaner

__all__ = ["load_table"]


def load_table(
    table_name: str, data_source: str = "", frequency: str = ""
) -> pd.DataFrame:
    name = ".".join([data_source, table_name, frequency])
    return ExternalDataCleaner(name).load_data()



import pandas as pd


def create_year_month_index(from_year: int, to_year: int) -> pd.MultiIndex:
    to_year += 1
    return pd.MultiIndex.from_frame(
        pd.concat(
            [
                pd.Series(range(from_year, to_year))
                .repeat(12)
                .reset_index(drop=True)
                .rename("Year"),
                pd.Series(list(range(1, 13)) * (to_year - from_year)).rename("Month"),
            ],
            axis="columns",
        )
    )


def cpi_by1395_urban_singleindex(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[2:, [2]]
    table.index = create_year_month_index(1361, 1401)
    table.columns = ["CPI"]
    return table


def cpi_by1395_rural_maingroups(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[3], 53:].T
    table.index = create_year_month_index(1374, 1401)
    table.columns = ["CPI"]
    return table


def cpi_by1395(table_list: list[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(
        table_list, keys=["Urban", "Rural"], names=["Urban_Rural", "Year", "Month"]
    )



import pandas as pd
import geopandas as gpd

from .metadata_reader import metadata, defaults
from . import utils


def create_geoseries(area: str, map_name: str) -> gpd.GeoSeries:
    file_name = metadata.maps[map_name][area]["file_name"]
    path = defaults.maps.joinpath(map_name, file_name)
    column_name = metadata.maps[map_name][area]["code_column"]
    codes = metadata.maps[map_name][area]["code"]
    area_names = metadata.household[area]["names"]
    codes = {value: area_names[key] for key, value in codes.items()}
    if not path.exists():
        utils.download_map(map_name)
    geoseries = (
        gpd.read_file(path)[[column_name, "geometry"]]
        .join(pd.Series(codes, name=area), on=column_name)
        .set_index(area)
        .loc[:, "geometry"]
    )
    return geoseries


def add_geometry(
    table: pd.DataFrame, area: str | None = None, map_name: str = "humandata"
) -> gpd.GeoDataFrame:
    names = table.index.names + table.columns.to_list()
    if area == "Region":
        assert "Region" in names
    elif area == "Province":
        assert "Province" in names
    elif "Region" in names:
        area = "Region"
    elif "Province" in names:
        area = "Province"
    else:
        raise ValueError

    geoseries = create_geoseries(area, map_name)
    geotable = gpd.GeoDataFrame(table.join(geoseries, on=area))
    return geotable



from typing import Literal, Iterable

import pandas as pd
from pydantic import BaseModel, Field

from .metadata_reader import Attribute as _Attribute
from . import api, utils


_QuantileBase = Literal[
    "Income",
    "Expenditure",
    "Gross_Expenditure",
    "Net_Expenditure",
]


_EquivalenceScale = Literal[
    "Household", "Per_Capita", "OECD", "OECD_Modified", "Square_Root"
]


class QuantileSettings(BaseModel):
    on_variable: _QuantileBase | None = Field(default=None, alias="on")
    on_column: str | None = None
    weighted: bool = True
    weight_column: str | None = None
    equivalence_scale: _EquivalenceScale = "Household"
    for_all: bool = True
    annual: bool = True
    groupby: _Attribute | Iterable[_Attribute] | None = None
    years: int | Iterable[int] | str | None = None

    def model_post_init(self, __context=None) -> None:
        if self.groupby is None:
            self.groupby = []
        if isinstance(self.groupby, str):
            self.groupby = [self.groupby]
        else:
            self.groupby = list(self.groupby)

        super().model_post_init(__context)


class Quantiler:
    """
    Quantiler Class
    """

    variable_aliases = {
        "Expenditure": "Gross_Expenditure",
    }

    variable_tables = {
        "Income": "Total_Income",
        "Gross_Expenditure": "Total_Expenditure",
        "Net_Expenditure": "Total_Expenditure",
    }

    def __init__(
        self,
        table: pd.DataFrame | pd.Series | None = None,
        settings: QuantileSettings | None = None,
    ) -> None:
        if settings is None:
            self.settings = QuantileSettings()
        else:
            self.settings = settings.model_copy()

        self.table = table
        self.years = self._find_years()

        if (self.settings.on_column is not None) and (self.table is not None):
            self.value_table = self._extract_value_table(self.table)
        elif self.settings.on_variable is not None:
            self.value_table = self._get_external_value_table(self.settings.on_variable)
        else:
            raise ValueError

    def _find_years(self) -> list[int]:
        if self.settings.years is not None:
            return utils.parse_years(self.settings.years)
        if self.table is None:
            raise ValueError("Year is Not available")
        if "Year" in self.table.index.names:
            return self.table.index.get_level_values("Year").unique().to_list()
        if "Year" in self.table.columns:
            return list(self.table["Year"].unique())
        raise ValueError("Year must be provided")

    def _extract_value_table(self, table: pd.DataFrame | pd.Series) -> pd.DataFrame:
        if isinstance(table, pd.Series):
            table = table.to_frame(name="Values")
            assert ("Year" in table.index.names) and ("ID" in table.index.names)
        if isinstance(table, pd.DataFrame):
            table = table.reset_index().set_index(["Year", "ID"])
            if len(table.columns) > 1:
                if self.settings.on_column is None:
                    raise ValueError("Column name must be provided")
                assert self.settings.on_column in table.columns
                table = table[[self.settings.on_column]]
            table.columns = ["Values"]
        return table

    def _get_external_value_table(self, variable) -> pd.DataFrame:
        variable = self.variable_aliases.get(variable, variable)
        table_name = self.variable_tables[variable]
        value_table = (
            api.load_table(table_name, self.years)
            .set_index(["Year", "ID"])[[variable]]
            .rename(columns={variable: "Values"})
        )
        if (not self.settings.for_all) and (self.table is not None):
            value_table = value_table.loc[self.table.set_index(["Year", "ID"]).index]
        return value_table

    def calculate_simple_quantile(self) -> pd.Series:
        groupby_columns: list = self.settings.groupby.copy()  # type: ignore
        if self.settings.annual:
            groupby_columns.append("Year")
        quantile = (
            self.value_table.dropna()
            .sort_values("Values")
            .pipe(self._add_attributes)
            .pipe(self._add_weights)
            .groupby(groupby_columns)
            .apply(self._calculate_subgroup_quantile)
        )
        if isinstance(quantile, pd.Series):
            quantile.index = pd.MultiIndex.from_arrays(
                [
                    quantile.index.get_level_values(-2),
                    quantile.index.get_level_values(-1),
                ]
            )
        else:
            quantile = quantile.iloc[0].rename("Quantile")
        return quantile

    def _calculate_subgroup_quantile(self, subgroup: pd.DataFrame) -> pd.Series:
        return subgroup.assign(
            CumWeight=lambda df: df["Weight"].cumsum(),
            Quantile=lambda df: df["CumWeight"] / df["CumWeight"].iloc[-1],
        ).loc[:, "Quantile"]

    def _add_weights(self, table: pd.DataFrame) -> pd.DataFrame:
        if self.settings.weighted:
            return api.add_weight(table)
        return table.assign(Weight=1)

    def _add_attributes(self, table: pd.DataFrame) -> pd.DataFrame:
        for attribute in self.settings.groupby:  # type: ignore
            table = api.add_attribute(table, attribute)  # type: ignore
        return table

    def calculate_quantile(self) -> pd.Series:
        equivalence_scale = (
            api.load_table("Equivalence_Scale", years=self.years)
            .set_index(["Year", "ID"])
            .loc[:, self.settings.equivalence_scale]
        )
        new_values = self.value_table["Values"].div(equivalence_scale)
        self.value_table.loc[:, "Values"] = new_values
        quantile = self.calculate_simple_quantile()
        return quantile


# pylint: disable=too-many-arguments
# pylint: disable=unused-argument
def calculate_quantile(
    *,
    table: pd.DataFrame | pd.Series | None = None,
    on: _QuantileBase | None = "Gross_Expenditure",
    on_column: str | None = None,
    weighted: bool = True,
    weight_column: str | None = None,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: _Attribute | Iterable[_Attribute] | None = None,
    years: int | Iterable[int] | str | None = None,
):
    settings_vars = {key: value for key, value in locals().items() if key != "table"}
    settings = QuantileSettings(**settings_vars)
    return Quantiler(table=table, settings=settings).calculate_quantile()


def calculate_decile(
    *,
    table: pd.DataFrame | pd.Series | None = None,
    on: _QuantileBase | None = "Gross_Expenditure",
    on_column: str | None = None,
    weighted: bool = True,
    weight_column: str | None = None,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: _Attribute | Iterable[_Attribute] | None = None,
    years: int | Iterable[int] | str | None = None,
):
    settings_vars = {key: value for key, value in locals().items() if key != "table"}
    settings = QuantileSettings(**settings_vars)
    quantile = Quantiler(table=table, settings=settings).calculate_quantile()
    return (
        quantile.multiply(10)
        .floordiv(1)
        .add(1)
        .clip(1, 10)
        .astype(int)
        .rename("Decile")
    )


def calculate_percentile(
    *,
    table: pd.DataFrame | pd.Series | None = None,
    on: _QuantileBase | None = "Gross_Expenditure",
    on_column: str | None = None,
    weighted: bool = True,
    weight_column: str | None = None,
    equivalence_scale: _EquivalenceScale = "Household",
    for_all: bool = True,
    annual: bool = True,
    groupby: _Attribute | Iterable[_Attribute] | None = None,
    years: int | Iterable[int] | str | None = None,
):
    settings_vars = {key: value for key, value in locals().items() if key != "table"}
    settings = QuantileSettings(**settings_vars)
    quantile = Quantiler(table=table, settings=settings).calculate_quantile()
    return (
        quantile.multiply(100)
        .floordiv(1)
        .add(1)
        .clip(1, 100)
        .astype(int)
        .rename("Percentile")
    )



"""
DataFrame extention
"""
from importlib import import_module

import pandas as pd

from .metadata_reader import Attribute as _Attributes
from . import decoder


@pd.api.extensions.register_dataframe_accessor("view")
class ViewAccessor:
    def __init__(self, pandas_obj: pd.DataFrame):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self._views = None

    @staticmethod
    def _validate(obj):
        pass

    @property
    def views(self) -> list[str] | None:
        return self._views

    @views.setter
    def views(self, value: list[str]):
        self._views = value

    def __getitem__(self, value: str) -> pd.DataFrame:
        if self._views is not None and value in self._views:
            settings = decoder.CommodityDecoderSettings(classification_name=value)
            return decoder.CommodityDecoder(self._obj, settings).add_classification()
        raise KeyError(f"{value} is not a view of the current table")


@pd.api.extensions.register_dataframe_accessor("hbsir")
class HBSIRFunctions:
    def __init__(self, pandas_obj: pd.DataFrame) -> None:
        self._obj = pandas_obj

    def add_classification(self, **kwargs):
        func = getattr(import_module("hbsir"), "add_classification")
        return func(self._obj, **kwargs)

    def add_attribute(self, **kwargs):
        func = getattr(import_module("hbsir"), "add_attribute")
        return func(self._obj, **kwargs)

    def add_weight(self, **kwargs):
        func = getattr(import_module("hbsir"), "add_weight")
        return func(self._obj, **kwargs)



import subprocess
from pathlib import Path
import platform

from ..metadata_reader import defaults
from .download_utils import download_7zip


def extract(compressed_file_path: str | Path, output_directory: str | Path) -> None:
    """
    Extracts the contents of a compressed file using the 7-Zip tool.

    :param compressed_file_path: The path to the compressed file to be extracted.
    :type compressed_file_path: str
    :param output_directory: The path to the directory where the extracted files will be stored.
    :type output_directory: str
    :return: None, as this function does not return anything.

    """
    if platform.system() == "Windows":
        _windows_extract(compressed_file_path, output_directory)
    elif platform.system() == "Linux":
        _linux_extract(compressed_file_path, output_directory)
    else:
        print("Your OS is not supported. try buy a new one.")


def _windows_extract(compressed_file_path: str | Path, output_directory: str | Path):
    seven_zip_file_path = defaults.package_dir.joinpath("7-Zip", "7z.exe")
    if not seven_zip_file_path.exists():
        download_7zip()
    subprocess.run(
        [
            seven_zip_file_path,
            "e",
            compressed_file_path,
            f"-o{output_directory}",
            "-y",
        ],
        check=False,
        shell=True,
    )


def _linux_extract(compressed_file_path: str | Path, output_directory: str | Path):
    seven_zip_file_path = defaults.package_dir.joinpath("7-Zip", "7zz")
    if not seven_zip_file_path.exists():
        download_7zip()
    subprocess.run(
        [
            seven_zip_file_path,
            "e",
            compressed_file_path,
            f"-o{output_directory}",
            "-y",
        ],
        check=False,
    )



from .seven_zip_utils import extract as sz_extract
from .download_utils import download, download_map
from .parsing_utils import parse_years, construct_table_year_pairs
from .metadata_utils import MetadataVersionResolver, MetadataCategoryResolver
from .argham import Argham



from itertools import product
from typing import Callable

import pandas as pd
from pydantic import BaseModel, Field

from . import utils
from .metadata_reader import metadata, defaults, Attribute as _Attribute


def read_classification_info(name, year):
    versioned_info = metadata.commodities[name]
    category_resolver = utils.MetadataCategoryResolver(versioned_info, year)
    classification_info = category_resolver.categorize_metadata()
    return classification_info


def create_classification_table(name, years) -> pd.DataFrame:
    table_list = []
    for year in years:
        classification_info = read_classification_info(name, year)
        annual_table = _create_annual_classification_table(classification_info)
        # pylint: disable=unsupported-assignment-operation
        annual_table["Year"] = year
        table_list.append(annual_table)
    table = pd.concat(table_list, ignore_index=True)
    return table


def _create_annual_classification_table(classification_info) -> pd.DataFrame:
    # pylint: disable=unsubscriptable-object
    table = pd.DataFrame(classification_info["items"])
    table["code_range"] = table["code"].apply(
        utils.Argham,  # type: ignore
        default_start=defaults.first_year,
        default_end=defaults.last_year + 1,
        keywords=["code"],
    )
    table = table.drop(columns=["code"])
    return table


def extract_column(table: pd.DataFrame, column_name: str) -> pd.Series:
    if column_name in table.columns:
        column = table.loc[:, column_name].copy()
    elif isinstance(table.index, pd.Index) and table.index.name == column_name:
        column = table.index.to_series()
    elif isinstance(table.index, pd.MultiIndex) and column_name in table.index.names:
        column = table.index.to_frame().loc[:, column_name].copy()
    else:
        raise KeyError
    return column


class CommodityDecoderSettings(BaseModel):
    name: str = Field(default="original", alias="classification_name")
    code_column_name: str = "Code"
    year_column_name: str = "Year"
    versioned_info: dict = {}
    defaults: dict = {}
    labels: tuple[str, ...] = tuple()
    levels: tuple[int, ...] = tuple()
    drop_value: bool = False
    output_column_names: tuple[str, ...] = ()
    required_columns: tuple[str, ...] | None = None
    missing_value_replacements: dict[str, str] | None = None

    def model_post_init(self, __contex=None) -> None:
        self.versioned_info = metadata.commodities[self.name]
        if "defaults" in self.versioned_info:
            self.defaults = self.versioned_info["defaults"]
        for key, value in self.defaults.items():
            if isinstance(value, list):
                value = tuple(value)
            if (getattr(self, key) is None) or (len(getattr(self, key)) == 0):
                setattr(self, key, value)
        if len(self.labels) == 0:
            self.labels = ("item_key",)
        if len(self.levels) == 0:
            self.levels = (1,)
        self._resolve_column_names()
        super().model_post_init(None)

    def _resolve_column_names(self) -> None:
        if len(self.output_column_names) == 0:
            names = [
                f"{label}_{level}" for label, level in product(self.labels, self.levels)
            ]
            self.output_column_names = tuple(names)
        elif len(self.output_column_names) == len(self.labels) * len(self.levels):
            pass
        elif len(self.output_column_names) == len(self.labels):
            names = [
                f"{label}_{level}"
                for label, level in product(self.output_column_names, self.levels)
            ]
            self.output_column_names = tuple(names)

    @property
    def rename_dict(self):
        return dict(zip(product(self.labels, self.levels), self.output_column_names))  # type: ignore


class CommodityDecoder:
    def __init__(self, table: pd.DataFrame, settings: CommodityDecoderSettings) -> None:
        self.table = table
        self.settings = settings
        self.code_column = extract_column(table, settings.code_column_name)
        self.year_column = extract_column(table, settings.year_column_name)
        self.classification_table = create_classification_table(
            name=self.settings.name,
            years=self.year_column.drop_duplicates().to_list(),
        )
        self.year_code_pairs = self.create_year_code_pairs()

    def create_year_code_pairs(self) -> pd.DataFrame:
        years = self.year_column.drop_duplicates()
        yc_pair_list = []
        for year in years:
            filt = self.year_column == year
            codes = self.code_column.loc[filt].drop_duplicates()
            yc_pair = codes.to_frame()
            yc_pair["Year"] = year
            yc_pair_list.append(yc_pair)
        return pd.concat(yc_pair_list, ignore_index=True)

    @staticmethod
    def _build_year_code_table(
        year_code_pairs: pd.DataFrame, row: pd.Series
    ) -> pd.DataFrame:
        filt = year_code_pairs["Code"].apply(lambda x: x in row["code_range"])
        filt = filt & (year_code_pairs["Year"] == row["Year"])
        matched_codes = year_code_pairs.loc[filt].set_index(["Year", "Code"])
        columns = row.drop(["code_range", "Year"]).index
        code_table = pd.DataFrame(
            data=[row.loc[columns]] * len(matched_codes.index),
            index=matched_codes.index,
            columns=columns,
        )
        return code_table

    def create_mapping_table(self) -> pd.DataFrame:
        code_table_list = []
        for _, row in self.classification_table.iterrows():
            code_table = self._build_year_code_table(self.year_code_pairs, row)
            if not code_table.empty:
                code_table_list.append(code_table)
        mapping_table = pd.concat(code_table_list)
        mapping_table = mapping_table.set_index("level", append=True)
        self._validate_mapping_table(mapping_table)
        mapping_table = mapping_table.unstack(-1)
        mapping_table = mapping_table.loc[:, self.settings.rename_dict.keys()]  # type: ignore
        mapping_table.columns = self.settings.rename_dict.values()
        return mapping_table

    @staticmethod
    def _validate_mapping_table(mapping_table: pd.DataFrame):
        filt = mapping_table.index.duplicated(keep=False)
        invalid_case_sample = (
            mapping_table.loc[filt].sort_values(["Code", "level"]).head(10)
        )
        if filt.sum() > 0:
            raise ValueError(f"Classification is not valid \n{invalid_case_sample}")

    def _fill_missing_values(self):
        if "missing_value_replacements" not in self.settings.defaults:
            return
        for column, default in self.settings.defaults[
            "missing_value_replacements"
        ].items():
            if column not in self.table.columns:
                continue
            filt = self.table.loc[:, column].isna()
            self.table.loc[filt, column] = default  # type: ignore

    def add_classification(self):
        mapping = self.create_mapping_table()
        self.table = self.table.join(mapping, on=["Year", "Code"])
        self._fill_missing_values()
        return self.table


class IDDecoderSettings(BaseModel):
    name: _Attribute = Field(alias="attribute_name")
    id_column_name: str = "ID"
    year_column_name: str = "Year"
    labels: tuple[str, ...] = ("names",)
    output_column_names: tuple[str] = tuple()

    def model_post_init(self, __contex=None) -> None:
        self._resolve_output_column_names()
        super().model_post_init(None)

    def _resolve_output_column_names(self) -> None:
        if len(self.output_column_names) != len(self.labels):
            if len(self.labels) == 1:
                names = [self.name]
            else:
                names = [f"{self.name}_{label}" for label in self.labels]
            self.output_column_names = tuple(names)


class IDDecoder:
    def __init__(
        self,
        table: pd.DataFrame,
        settings: IDDecoderSettings,
    ) -> None:
        self.table = table
        self.settings = settings
        self.id_column = extract_column(table, settings.id_column_name)
        self.year_column = extract_column(table, settings.year_column_name)

    def construct_mapping_table(self):
        mapped_columns = [self.year_column, self.id_column]
        for label in self.settings.labels:
            mapped_column = self.map_id_to_label(label)
            mapped_columns.append(mapped_column)
        year_and_id = [self.settings.year_column_name, self.settings.id_column_name]
        columns = year_and_id + list(self.settings.output_column_names)
        mapping_table = pd.concat(mapped_columns, axis="columns", keys=columns)
        mapping_table = mapping_table.drop_duplicates().set_index(year_and_id)
        return mapping_table

    def _get_metadata_version(self, year) -> dict:
        household_metadata = utils.MetadataVersionResolver(
            metadata.household, year
        ).get_version()
        assert isinstance(household_metadata, dict)
        return household_metadata

    def _create_code_builder(self, household_metadata: dict) -> Callable:
        ld_len = household_metadata["ID_Length"]
        attr_dict = household_metadata[self.settings.name]
        if ("position" not in attr_dict) or attr_dict["position"] is None:
            raise ValueError("Code position is not available")
        start, end = attr_dict["position"]["start"], attr_dict["position"]["end"]

        def builder(household_id_column: pd.Series) -> pd.Series:
            return (
                household_id_column
                % pow(10, (ld_len - start))
                // pow(10, (ld_len - end))
            )

        return builder

    def _create_code_mapper(self, label, year) -> Callable:
        if label == "code":
            return self._create_code_builder

        household_metadata = self._get_metadata_version(year)
        # pylint: disable=unsubscriptable-object
        mapping = household_metadata[self.settings.name][label]
        code_builder = self._create_code_builder(household_metadata)

        def mapper(household_id_column: pd.Series) -> pd.Series:
            mapped = code_builder(household_id_column).map(mapping).astype("category")
            mapped.name = label
            return mapped

        return mapper

    def map_id_to_label(self, label):
        years = self.year_column.drop_duplicates()
        attribute_column = pd.Series(index=self.table.index, dtype="object")
        for year in years:
            filt = self.year_column == year
            attribute_column.loc[filt] = self._create_code_mapper(label, year)(
                self.id_column.loc[filt]
            )
        return attribute_column

    def add_attribute(self):
        mapping_table = self.construct_mapping_table()
        year_and_id = [self.settings.year_column_name, self.settings.id_column_name]
        self.table = self.table.join(mapping_table, year_and_id)
        return self.table
